#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <json-c/json_tokener.h>
#include <sys/stat.h>

#include "config.h"
#include "netex.h"
#include "strutil.h"

#define DELIM_CHAR '.'

const char* config_to_formatted_json_string(const CONFIGURATION* config)
{
    if (config == NULL)
        return NULL;
    return json_object_to_json_string_ext(config->json, JSON_C_TO_STRING_PRETTY | JSON_C_TO_STRING_SPACED);
}
CONFIGURATION* config_create(void)
{
    CONFIGURATION* config_obj = malloc(sizeof(CONFIGURATION));
    config_obj->json = json_object_new_object();
    return config_obj;
}
int config_clear(const CONFIGURATION* config)
{
    if (config == NULL)
        return -1;
    if (config->json != NULL)
        json_object_put(config->json);
    return 0;
}
CONFIGURATION* config_load_from_path(const char* path)
{
    if (path == NULL)
        return NULL;
    struct stat con_file_stat;
    if (lstat(path, &con_file_stat) == -1)
    {
        const int error = errno;
        WARN("Failed to get stat from file '%s'. Error code: %i", path, error);
        return NULL;
    }
    const long int file_size = con_file_stat.st_size;
    if (file_size <= 0)
    {
        //TODO: Empty file?
        return NULL;
    }
    FILE* config_file = fopen(path, "r");
    if (config_file == NULL)
    {
        WARN("Failed to load file '%s'", path);
        return NULL;
    }
    void* config_mem = malloc(file_size);
    if (config_mem == NULL)
    {
        const int error = errno;
        WARN("Failed to allocate %ld bytes, error: %i", file_size, error);
        return NULL;
    }
    const size_t total_read = fread(config_mem, file_size, file_size, config_file);
    fclose(config_file);
    if (total_read != 1)
    {
        free(config_mem);
        WARN("Failed to load read file!");
        return NULL;
    }
    json_object* tokenized = json_tokener_parse(config_mem);
    free(config_mem);
    if (tokenized == NULL)
    {
        WARN("Failed to tokenize JSON!");
        return NULL;
    }
    CONFIGURATION* config_obj = malloc(sizeof(CONFIGURATION));
    if (config_obj == NULL)
    {
        WARN("Failed to allocate memory!");
        return NULL;
    }
    config_obj->json = tokenized;
    return config_obj;
}

int config_save_to_path(const CONFIGURATION* config, const char* path)
{
    if (config == NULL || path == NULL)
        return -1;
    FILE* config_file = fopen(path, "w");
    if (config_file == NULL)
    {
        const int error = errno;
        WARN("Failed to open file: '%s' Error code: %i", path, error);
        return error;
    }
    const char* json_str = json_object_to_json_string_ext(config->json, JSON_C_TO_STRING_PRETTY | JSON_C_TO_STRING_SPACED);
    if (json_str == NULL)
    {
        WARN("Failed to convert JSON obj to string.");
        return -1;
    }
    const size_t json_str_len = strlen(json_str);
    const int write_result = fwrite(json_str, 1, json_str_len, config_file);
    fclose(config_file);
    if (write_result <= 0)
    {
        const int wr_error = errno;
        WARN("Error occurred while writing file '%s' to disk, error code: %i", path, wr_error);
        return wr_error;
    }
    return 0;
}
int config_remove(const CONFIGURATION* config, const char* key)
{
    if (key == NULL)
        return -1;
    json_object* to_remove = config_get_json(config, key);
    if (json_object_put(to_remove) != 1)
    {
        WARN("Failed to remove item from JSON");
        return -1;
    }
    return 0;
}
int config_set_string(const CONFIGURATION* config, const char* key, const char* str_value)
{
    if (str_value == NULL)
        return -1;
    json_object* str_json_value = json_object_new_string(str_value);
    return config_set_json(config, key, str_json_value);
}
int config_set_int(const CONFIGURATION* config, const char* key, const size_t int_value)
{
    json_object* int_value_json = json_object_new_int64(int_value);
    return config_set_json(config, key, int_value_json);
}
int config_set_uint(const CONFIGURATION* config, const char* key, const uint64_t int_value)
{
    json_object* int_value_json = json_object_new_uint64(int_value);
    return config_set_json(config, key, int_value_json);
}
int config_set_double(const CONFIGURATION* config, const char* key, const double double_value)
{
    json_object* double_value_json = json_object_new_double(double_value);
    return config_set_json(config, key, double_value_json);
}

int config_set_json(const CONFIGURATION* config, const char* key, json_object* json_value_object)
{
    if ((config == NULL || json_value_object == NULL) || config->json == NULL)
        return -1;
    SPLIT_STR* splitted_key = splstr(key, DELIM_CHAR);
    if (splitted_key == NULL || splitted_key->count == 0)
        return -1;
    json_object* temp_obj = NULL;
    for (int i = 0; i != splitted_key->count; i++)
    {
        json_object* current_obj;
        if (temp_obj == NULL)
            current_obj = config->json;
        else
            current_obj = temp_obj;

        // If object with the key already exists, set the temp variable to the found object and continue the loop. (The found object will be used next time in the loop.)
        if (json_object_object_get_ex(current_obj, splitted_key->delimited[i], &temp_obj))
            continue;

        if (i == splitted_key->count - 1) // Last key (set value)
        {
            if (json_object_object_add(current_obj, splitted_key->delimited[i], json_value_object) != 0)
                ERROR("Failed to set value to JSON!");
            break;
        }

        json_object* new_json = json_object_new_object();
        if (json_object_object_add(current_obj, splitted_key->delimited[i], new_json) != 0)
        {
            ERROR("Failed to set new JSON object!");
            break;
        }
        temp_obj = new_json;
    }
    strsplfree(splitted_key);
    return 0;
}

char* config_get_string(const CONFIGURATION* config, const char* key)
{
    json_object* json_string_obj = config_get_json(config, key);
    if (json_string_obj == NULL)
        return NULL;
    // Duplicate the string, if the json object is free'd (using json_object_put()) the returned string is still there and can be used further in the application.
    return strdup(json_object_get_string(json_string_obj));
}
int64_t config_get_int(const CONFIGURATION* config, const char* key)
{
    const json_object* json_int_obj = config_get_json(config, key);
    if (json_int_obj == NULL)
        return 0;
    return json_object_get_int64(json_int_obj);
}
uint64_t config_get_uint(const CONFIGURATION* config, const char* key)
{
    const json_object* json_int_obj = config_get_json(config, key);
    if (json_int_obj == NULL)
        return 0;
    return json_object_get_uint64(json_int_obj);
}
double config_get_double(const CONFIGURATION* config, const char* key)
{
    const json_object* json_double_obj = config_get_json(config, key);
    if (json_double_obj == NULL)
        return 0;
    return json_object_get_double(json_double_obj);
}

json_object* config_get_json(const CONFIGURATION* config, const char* key)
{
    if (config == NULL || config->json == NULL)
        return NULL;
    SPLIT_STR* splitted = splstr(key, DELIM_CHAR);
    if (splitted == NULL || splitted->count == 0)
        return NULL;
    json_object* tmp_json = config->json;
    for (int i = 0; i != splitted->count; i++)
    {
        json_object* current_json_object = json_object_new_object();
        if (json_object_object_get_ex(tmp_json, splitted->delimited[i], &current_json_object))
            tmp_json = current_json_object;
        else // No key found
        {
            strsplfree(splitted);
            return NULL;
        }
    }
    strsplfree(splitted);
    return tmp_json;
}